import argparse
import asyncio
import aiofiles
import sys
from aioconsole.stream import create_standard_streams
from exitstatus import ExitStatus

from asyncfileserver import __version__

from asyncfileserver.infra.file import File
from asyncfileserver.model.client import Client
from asyncfileserver.infra.async_console_input import AsyncConsoleInput
from asyncfileserver.infra.async_console_output import AsyncConsoleOutput
from asyncfileserver.model.ask_answer_arbiter import AskAnswerArbiter as Arbiter
from asyncfileserver.model.confirm_put_queue import ConfirmPutQueue
from asyncfileserver.model.view_data_factory import ViewDataFactory
from asyncfileserver.model.confirm_command_factory import ConfirmCommandFactory


async def asyncfileserver(file_name: str) -> int:
    async with aiofiles.open(file_name, "rb") as async_file:
        streams = await create_standard_streams(sys.stdin.buffer,
                                                sys.stdout.buffer,
                                                sys.stderr.buffer)
        reader, writer, _ = streams
        input = AsyncConsoleInput(reader)
        output = AsyncConsoleOutput(writer)
        arbiter = Arbiter(input, output, ViewDataFactory(),
                          ConfirmCommandFactory())
        queue = ConfirmPutQueue(arbiter, asyncio.Queue())
        file = File(file=async_file, queue=queue)

        client = Client(queue, output)

        read_file = asyncio.create_task(file.read())
        write_console = asyncio.create_task(client.write())

        await asyncio.gather(read_file, write_console)

        return ExitStatus.success


def main():

    parser = argparse.ArgumentParser(description="Start a file server.")

    parser.add_argument("-f", "--file", type=str,
                        help="file to be served")

    parser.add_argument("-v", "--version", action='store_true',
                        help="show version")

    args = parser.parse_args()

    if args.version:
        print(f"asyncfileserver version {__version__}")
        sys.exit(ExitStatus.success)

    if args.file == None:
        parser.error("the following argument is required: FILE")

    loop = asyncio.get_event_loop()

    try:
        status = loop.run_until_complete(asyncfileserver(args.file))
    except FileNotFoundError as e:
        parser.error(f'file "{args.file}": not found')
    except TypeError as e:
        parser.error(f'type error: {e}')
    finally:
        loop.close()

    sys.exit(status)


if __name__ == "__main__":
    main()



class Client(object):
    def __init__(self, queue, output):
        self._queue = queue
        self._output = output

    async def write(self):
        item = await self._queue.get()
        while(item):
            await self._output.print(item)
            self._queue.task_done()
            item = await self._queue.get()

        self._queue.task_done()



import asyncio
import aiounittest

from asyncfileserver.infra.file import File
from asyncfileserver.tests.unit.model.fake_async_queue import FakeAsyncQueue


class ByteArrayFile(object):
    def __init__(self, buffer: bytearray):
        self._buffer = buffer
        self._index = 0

    async def read(self, size):
        read_size = min(size, len(self._buffer[self._index:]))
        end_index = self._index + read_size
        data = self._buffer[self._index:end_index]
        self._index = self._index + read_size
        return data


class TestFile(aiounittest.AsyncTestCase):

    def get_event_loop(self):
        return asyncio.get_event_loop()

    async def test_empty_file(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"")), queue = async_queue)

        data = file.data()

        with self.assertRaises(StopAsyncIteration):
            await data.__anext__()

    async def test_read_buffer_size_lt_file_size_lt_queue_item_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=1, queue_item_size=6)
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == [bytearray(b"\x01\x02\x03\x04\x05")])

    async def test_read_buffer_size_lt_file_size_eq_queue_item_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=1, queue_item_size=5)
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == [bytearray(b"\x01\x02\x03\x04\x05")])

    async def test_read_buffer_size_lt_file_size_gt_queue_item_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=1, queue_item_size=4)
        result = [
            bytearray(b"\x01\x02\x03\x04"),
            bytearray(b"\x05")
        ]

        items = []

        async for item in file.data():
            items.append(item)

        self.assertTrue(items == result)

    async def test_read_buffer_size_eq_queue_item_size_lt_file_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=1, queue_item_size=1)
        result = [bytearray([i+1]) for i in range(5)]

        items = []

        async for item in file.data():
            items.append(item)

        self.assertTrue(items == result)

    async def test_file_size_lt_read_buffer_size_lt_queue_item_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=6, queue_item_size=7)
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == [bytearray(b"\x01\x02\x03\x04\x05")])

    async def test_file_size_lt_read_buffer_size_eq_queue_item_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=6, queue_item_size=6)
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == [bytearray(b"\x01\x02\x03\x04\x05")])

    async def test_file_size_lt_queue_item_size_lt_read_buffer_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=7, queue_item_size=6)
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == [bytearray(b"\x01\x02\x03\x04\x05")])

    async def test_file_size_eq_queue_item_size_lt_read_buffer_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=6, queue_item_size=5)
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == [bytearray(b"\x01\x02\x03\x04\x05")])

    async def test_queue_item_size_lt_file_size_lt_read_buffer_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=4, queue_item_size=1)
        result = [bytearray([i+1]) for i in range(5)]
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == result)

    async def test_queue_item_size_lt_file_size_eq_read_buffer_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=5, queue_item_size=1)
        result = [bytearray([i+1]) for i in range(5)]
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == result)

    async def test_queue_item_size_lt_read_buffer_size_lt_file_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=4, queue_item_size=1)
        result = [bytearray([i+1]) for i in range(5)]
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == result)

    async def test_queue_item_size_eq_read_buffer_size_lt_file_size(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=1, queue_item_size=1)
        result = [bytearray([i+1]) for i in range(5)]
        items = []
        async for item in file.data():
            items.append(item)

        self.assertTrue(items == result)

    async def test_read_without_consuming_queue(self):
        queue = []
        async_queue = FakeAsyncQueue(queue)
        file = File(ByteArrayFile(bytearray(b"\x01\x02\x03\x04\x05")),
                    queue=async_queue,
                    read_buffer_size=1, queue_item_size=1)

        result = [bytearray([i+1]) for i in range(5)]
        result.extend([None, None])

        await file.read()
        self.assertEqual(async_queue.how_many_tasks_done(), 0)

        async for item in file.data():
            pass
        self.assertEqual(async_queue.how_many_tasks_done(), 6)

        self.assertEqual(queue, result)



import asyncio
import aiounittest

from asyncfileserver.infra.async_console_input import AsyncConsoleInput


class MockReader(object):
    def __init__(self, lines):
        self._lines = lines
        self._index = 0

    async def readline(self):
        await asyncio.sleep(0)

        if self._index >= len(self._lines):
            return b''

        item = self._lines[self._index]
        self._index = self._index + 1
        return item


class TestAsyncConsoleInput(aiounittest.AsyncTestCase):

    def get_event_loop(self):
        return asyncio.get_event_loop()

    async def test_read_line_input(self):
        singular_object = object()
        lines = [singular_object]
        reader = MockReader(lines)
        console = AsyncConsoleInput(reader)

        self.assertEqual(await console.input(), singular_object)
        self.assertEqual(await console.input(), b'')



import unittest

from asyncfileserver.model.view_data_factory import ViewDataFactory


class TestViewDataFactory(unittest.TestCase):
    def test_create(self):
        factory = ViewDataFactory()
        data = bytearray(b'\x0135=X\x01')
        instance = factory.create(data)

        self.assertEqual(f"{instance}", f"<DATA size: {len(data)}>")



import re


class ConfirmCommand(object):
    _pattern = re.compile(b'\W')

    def __init__(self, data: bytearray):
        self._data = self._format(data)

    def go_on(self):
        return self._data == b"C"

    def yes(self):
        return self._data == b"Y"

    def _format(self, data):
        upper = data.upper()
        return re.sub(self._pattern, b'', upper) 



import asyncio
import aiounittest

from asyncfileserver.infra.async_console_output import AsyncConsoleOutput


class MockWriter(object):
    def __init__(self, buffer):
        self._buffer = buffer

    def write(self, msg):
        self._buffer.append(msg)

    async def drain(self):
        await asyncio.sleep(0)


class TestAsyncConsoleInput(aiounittest.AsyncTestCase):

    def get_event_loop(self):
        return asyncio.get_event_loop()

    async def test_write_line_output(self):
        lines = []
        writer = MockWriter(lines)
        console = AsyncConsoleOutput(writer)

        singular_object = object()
        await console.print(singular_object)
        self.assertEqual(lines, [singular_object])



import asyncio
import aiofiles


class File(object):

    def __init__(self, file,
                 read_buffer_size: int = 2048,
                 queue=asyncio.Queue(), queue_item_size: int = 91):
        self._file = file
        self._buffer_size = read_buffer_size
        self._queue = queue
        self._queue_item_size = queue_item_size
        self._buffer_index = 0
        self._buffer = bytearray()

    async def data(self):
        read = asyncio.create_task(self.read())
        item = await self._queue.get()
        while item:
            yield item
            self._queue.task_done()
            item = await self._queue.get()

        self._queue.task_done()
        await read

    async def read(self):
        while await self._read():
            while await self._append():
                pass
            self._buffer = self._buffer[self._buffer_index:]
            self._buffer_index = 0

        if self._remaining_buffer_size() > 0:
            await self._queue.put(self._buffer[self._buffer_index:])

        await self._queue.put(None)

    async def _read(self):
        chunk = await self._file.read(self._buffer_size)
        self._buffer.extend(chunk)
        return chunk

    def _remaining_buffer_size(self):
        return len(self._buffer) - self._buffer_index

    async def _append(self):
        if self._remaining_buffer_size() >= self._queue_item_size:
            end_index = self._buffer_index + self._queue_item_size
            item = bytearray(self._buffer[self._buffer_index:end_index])
            await self._queue.put(item)
            self._buffer_index = self._buffer_index + self._queue_item_size
            return True
        return False



import asyncio
import aiounittest

from asyncfileserver.model.confirm_put_queue import ConfirmPutQueue
from asyncfileserver.tests.unit.model.fake_async_queue import FakeAsyncQueue


class AllowAll(object):
    async def should_put(self, item):
        # Release control back to event loop to simulate more properly an async
        # call.
        await asyncio.sleep(0)

        return True


class AllowSome(object):
    def __init__(self, items):
        self._items = items
        self._index = 0

    async def should_put(self, item):
        await asyncio.sleep(0)

        if self._index >= len(self._items):
            return False

        if item == self._items[self._index]:
            self._index = self._index + 1
            return True

        return False


class TestConfirmPutQueue(aiounittest.AsyncTestCase):

    @staticmethod
    async def _compose_get_task_done(queue):
        item = await queue.get()
        queue.task_done()
        return item

    def get_event_loop(self):
        return asyncio.get_event_loop()

    async def test_allow_all_arbiter(self):
        allow_all = AllowAll()
        queue = []
        fake_queue = FakeAsyncQueue(queue)
        confirm_queue = ConfirmPutQueue(allow_all, fake_queue)

        singular_item = bytearray(b'')
        put_task = asyncio.create_task(confirm_queue.put(singular_item))
        get_task = asyncio.create_task(
            self._compose_get_task_done(confirm_queue))
        _, same_element = await asyncio.gather(put_task, get_task)

        self.assertEqual(same_element, b'')

    async def test_allow_some_elements(self):
        allowed_bytes = [bytes([i]) for i in (1, 3, 5, 7, 9)]
        allowed_bytearrays = [bytearray(i) for i in allowed_bytes]
        allow_some = AllowSome(allowed_bytearrays)
        queue = []
        fake_queue = FakeAsyncQueue(queue)
        confirm_queue = ConfirmPutQueue(allow_some, fake_queue)

        put_tasks = [asyncio.create_task(
            confirm_queue.put(bytearray([i]))) for i in range(10)]

        get_tasks = [asyncio.create_task(
            self._compose_get_task_done(confirm_queue)) for _ in range(5)]

        * results, = await asyncio.gather(* get_tasks, * put_tasks)

        self.assertEqual(results[0:5], allowed_bytes)



import asyncio


class QueueInput(object):
    def __init__(self, queue: list):
        self._queue = queue
        self._index = 0

    async def input(self):
        await asyncio.sleep(0)

        if self._index >= len(self._queue):
            return b''

        item = self._queue[self._index]
        self._index = self._index + 1
        return item



import asyncio
import aiounittest

from asyncfileserver.model.ask_answer_arbiter import AskAnswerArbiter
from asyncfileserver.tests.unit.infra.buffer_output import BufferOutput
from asyncfileserver.tests.unit.infra.queue_input import QueueInput


class ViewDataMock(object):
    def __init__(self, representation):
        self._representation = representation

    def __str__(self):
        return self._representation


class ConfirmCommandMock(object):
    def __init__(self, go_on, yes):
        self._go_on = go_on
        self._yes = yes

    def go_on(self):
        return self._go_on

    def yes(self):
        return self._yes


class FactoryMock(object):
    def __init__(self, instances):
        self._instances = instances
        self._index = 0

    def create(self, data: bytearray):
        instance = self._instances[self._index]
        self._index = self._index + 1
        return instance


class TestConsoleArbiter(aiounittest.AsyncTestCase):

    def get_event_loop(self):
        return asyncio.get_event_loop()

    async def test_simple_inquire_response(self):
        output_buffer = []
        output = BufferOutput(output_buffer)

        input_buffer = [object()]
        input_queue = QueueInput(input_buffer)

        view = ViewDataMock("DATA")
        command = ConfirmCommandMock(False, True)

        arbiter = AskAnswerArbiter(input_queue, output,
                                   FactoryMock([view]),
                                   FactoryMock([command]))

        singular_object = object()

        self.assertTrue(await arbiter.should_put(singular_object))
        self.assertEqual(output_buffer, [f"\n{view}\n> "])

    async def test_deny_and_allow(self):
        input_buffer = [object(), object(), object(), object()]
        input_queue = QueueInput(input_buffer)

        output_buffer = []
        output = BufferOutput(output_buffer)

        view1 = ViewDataMock("DATA1")
        view2 = ViewDataMock("DATA2")
        view3 = ViewDataMock("DATA3")
        view4 = ViewDataMock("DATA4")

        command1 = ConfirmCommandMock(False, False)
        command2 = ConfirmCommandMock(False, True)
        command3 = ConfirmCommandMock(True, False)
        command4 = ConfirmCommandMock(False, False)

        arbiter = AskAnswerArbiter(input_queue, output,
                                   FactoryMock([view1, view2, view3, view4]),
                                   FactoryMock([command1, command2, command3,
                                                command4]))

        singular_object1 = object()
        singular_object2 = object()
        singular_object3 = object()
        singular_object4 = object()

        self.assertFalse(await arbiter.should_put(singular_object1))
        self.assertEqual(output_buffer, [f"\n{view1}\n> "])

        self.assertTrue(await arbiter.should_put(singular_object2))
        self.assertEqual(output_buffer, [f"\n{view1}\n> ", f"\n{view2}\n> "])

        self.assertTrue(await arbiter.should_put(singular_object3))
        self.assertEqual(output_buffer,
                         [f"\n{view1}\n> ", f"\n{view2}\n> ", f"\n{view3}\n> "])

        self.assertTrue(await arbiter.should_put(singular_object4))
        self.assertEqual(output_buffer,
                         [f"\n{view1}\n> ", f"\n{view2}\n> ", f"\n{view3}\n> "])



class AskAnswerArbiter(object):
    def __init__(self, input, output, view_data_factory, confirm_command_factory):
        self._input = input
        self._output = output
        self._continue = False
        self._view_data_factory = view_data_factory
        self._confirm_command_factory = confirm_command_factory

    async def should_put(self, item):
        if self._continue:
            return True

        view_item = self._view_data_factory.create(item)
        await self._output.print(f"\n{view_item}\n> ")
        command = self._confirm_command_factory.create(await self._input.input())

        self._continue = command.go_on()
        return command.go_on() or command.yes()



import asyncio
import aiounittest
import aioconsole

from asyncfileserver.model.client import Client
from ..infra.buffer_output import BufferOutput


class FixedElementsQueue(object):
    def __init__(self, elements):
        self._elements = elements
        self._index = 0

    async def get(self):
        if self._index >= len(self._elements):
            return None

        element = self._elements[self._index]

        return element

    def task_done(self):
        self._index = self._index + 1

    def how_many_tasks_done(self) -> int:
        return self._index


class TestConsoleClient(aiounittest.AsyncTestCase):

    def get_event_loop(self):
        return asyncio.get_event_loop()

    async def test_simple_output(self):
        singular_element = object()
        queue = FixedElementsQueue([singular_element])

        output_elements = []
        output = BufferOutput(output_elements)

        client = Client(queue, output)

        await client.write()
        self.assertEqual(output_elements[0], singular_element)
        self.assertEqual(queue.how_many_tasks_done(), 2)



import asyncio
import aiounittest

from asyncfileserver.model.simple_queue import SimpleQueue
from ..model.fake_async_queue import FakeAsyncQueue


class TestSimpleQueue(aiounittest.AsyncTestCase):

    @staticmethod
    async def _compose_get_task_done(queue):
        item = await queue.get()
        queue.task_done()
        return item

    def get_event_loop(self):
        return asyncio.get_event_loop()

    async def test_single_empty_bytearray(self):
        queue = []
        fake_queue = FakeAsyncQueue(queue)
        confirm_queue = SimpleQueue(fake_queue)

        singular_item = bytearray(b'')
        put_task = asyncio.create_task(confirm_queue.put(singular_item))
        get_task = asyncio.create_task(
            self._compose_get_task_done(confirm_queue))
        _, same_element = await asyncio.gather(put_task, get_task)

        self.assertEqual(same_element, b'')

    async def test_get_and_put_simultaneously(self):
        queue = []
        fake_queue = FakeAsyncQueue(queue)
        confirm_queue = SimpleQueue(fake_queue)

        put_tasks = [asyncio.create_task(
            confirm_queue.put(bytearray([i]))) for i in range(10)]

        get_tasks = [asyncio.create_task(
            self._compose_get_task_done(confirm_queue)) for _ in range(10)]

        * results, = await asyncio.gather(* get_tasks, * put_tasks)

        self.assertEqual(results[0:len(get_tasks)], queue)



import unittest

from asyncfileserver.model.confirm_command_factory import ConfirmCommandFactory

class TestConfirmCommandFactory(unittest.TestCase):
    def test_create(self):
        factory = ConfirmCommandFactory()
        instance = factory.create(bytearray(b'Y\n'))

        self.assertFalse(instance.go_on())
        self.assertTrue(instance.yes())


class ConfirmPutQueue(object):
    def __init__(self, arbiter, queue):
        self._arbiter = arbiter
        self._queue = queue

    async def get(self) -> bytes:
        item: bytearray = await self._queue.get()
        return bytes(item) if item != None else None

    def task_done(self):
        self._queue.task_done()

    async def put(self, item: bytearray):
        if await self._arbiter.should_put(item):
            await self._queue.put(item)



import asyncio


class FakeAsyncQueue(object):
    def __init__(self, queue: list):
        self._queue = queue
        self._count_task_done = 0

    async def put(self, item):
        self._queue.append(item)
        await asyncio.sleep(0)

    async def get(self):
        await asyncio.sleep(0)
        await asyncio.sleep(0)
        return self._queue[self._count_task_done]

    def task_done(self):
        self._count_task_done = self._count_task_done + 1

    def how_many_tasks_done(self) -> int:
        return self._count_task_done
